/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * System Monitor
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : main.c
 * CREATION DATE : Apr 6, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <errno.h>
#include <stdlib.h>                    // for calloc, free, exit and its status
#include <getopt.h>                    // for options
#include <libgen.h>                    // for basename

#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include <nmapp/logging/SysLogger.h>
#include <nmapp/conversion/conversion.h>
#include <nmapp/file/FileOperation.h>
#include <nmapp/sys/UnixInitSys.h>
#include "sys-mon.h"


/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/

// Constants:
//------------------------
#define        DEFAULT_APPNAME                 "sys-mon"
#define        DEFAULT_MONITORING_TIMEOUT_MS    60000
#define        DEFAULT_LOGGING_TIMEOUT_MS      300000

#define        LOG_NAME_FILEOP                 "fileOp"
const char*    SYS_TEMP_FILEPATH             = "/sys/class/thermal/thermal_zone0/temp";

static const struct option APP_LONGOPTS[] =
{
    {"help",         no_argument,        NULL, 'h'},
    {"log-interval", required_argument,  NULL, 'l'},
    {"mon-interval", required_argument,  NULL, 'm'},
    {"daemonize",    no_argument,        NULL, 'd'},
    {"pidfile",      required_argument,  NULL, 'p'},
    SYSLOGGER_LONGOPTS,
    {NULL,        0,                  NULL,  0 }
};
static const char* APP_SHORTOPTS = "hl:m:dp:";

// Containers:
//------------------------
typedef struct __FileReader
{
    FileMonitor_t    def;
    FileOperationRef opRef;
} FileReader_t;


// Variables:
//------------------------
static const char*      myAppName        = DEFAULT_APPNAME;
static uint32_t         myTimeoutMS      = (uint32_t)(DEFAULT_MONITORING_TIMEOUT_MS);
static uint32_t         myLogIntervalMS  = (uint32_t)(DEFAULT_LOGGING_TIMEOUT_MS);
static bool             hasToDaemonize   = false;
static const char*      myPidFilePath    = NULL;
static SysLoggerRef     mySysLogger      = NULL;
static SysMonRef        mySysMonRef      = NULL;
static ILogRef          myFileOpLogRef   = NULL;
static FileReader_t     mySysTempReader  = {.def={NULL,NULL}, .opRef = NULL};
static const char*      myFileOpLogName  = (const char*)(LOG_NAME_FILEOP);


//--------------------------------------------------------------------------------------------
// Functions:
//------------------------
//--------------------------------------------------------------------------------------------
static void printAppUsage(void)
{
    printf("Usage: %s [args]\n\n"
           "    -h | --help                 Show this help\n"
           "    -l | --log-interval=timeMS  The logging interval [ms], default=%ums\n"
           "    -m | --mon-interval=timeMS  The monitoring interval [ms], default=%ums\n"
           "    -d | --daemonize            Run as daemon\n"
           "    -p | --pidfile=path         The PID file, see -d\n"
           "\n", myAppName, myLogIntervalMS, myTimeoutMS
    );
    const char* evtLoopLogName    = LOG_NAME_EVTLOOP;
    const char* tempSensorLogName = LOG_NAME_TEMPSENSOR;
    printf("Used loggers: - %s\n"
           "              - %s\n"
           "              - %s\n\n", evtLoopLogName, myFileOpLogName, tempSensorLogName);
    SysLogCmdLine_printHelp();
}
//--------------------------------------------------------------------------------------------
void handleParsedArgument(int parsedShortArg, int argc, char **argv, int optIdx)
{
    switch (parsedShortArg)
    {
        case 'h':
            printAppUsage();
            exit(EXIT_SUCCESS);
        case 'd':
            hasToDaemonize = true;
            break;
        case 'm':
            myTimeoutMS = str2uint(optarg);
            break;
        case 'l':
            myLogIntervalMS = str2uint(optarg);
            break;
        case 'p':
            myPidFilePath = (const char*)optarg;
            break;
        case '?':
        default:
        {
            if(!SysLogCmdLine_isValidCommandLineOption(argv[optind - 1]))
            {
                printAppUsage();
                exit(EXIT_FAILURE);
            }
        }
    } // end switch
}
//--------------------------------------------------------------------------------------------
SysLoggerRef parseAppArgumentsAndGetLoggerConfig(int argc, char **argv)
{
    Exception_t  ex;
    SysLoggerRef loggerRef = NULL;
    int c;
    int optIdx = 0;

    Try
    {
        while(true)
        {
            c = getopt_long(argc, argv, APP_SHORTOPTS, APP_LONGOPTS, &optIdx);
            if (c == -1)
            {
                break;
            }
            handleParsedArgument(c, argc, argv, optIdx);
        }

        loggerRef = SysLogger_create(myAppName, LOG_LEVEL_INFO);
        SysLoggersCmdLine_t loggerCfg = SysLogCmdLine_parse(argc, argv);
        SysLogger_addCommandLineLoggers(&loggerCfg);
    }
    Catch(ex)
    {
        fprintf(stderr, "Failed parsing application arguments: %s", ex.msg);
        exit(EXIT_FAILURE);
    }
    return loggerRef;
}
//--------------------------------------------------------------------------------------------
void buildingApp(void)
{
    ILogRef logRef = SysLogger_getLoggerItf(mySysLogger);
    ILog_msg(logRef, LOG_LEVEL_INFO, "%s starting with interval settings: monitoring=%ums, logging=%ums",
                                     myAppName, (unsigned int)myTimeoutMS, (unsigned int)myLogIntervalMS);

    myFileOpLogRef  = SysLogger_getLoggerInstance(myFileOpLogName, true);
    mySysTempReader.opRef         = FileOperation_create(myFileOpLogRef);
    mySysTempReader.def.fileName  = SYS_TEMP_FILEPATH;
    mySysTempReader.def.fileOpItf = FileOperation_getItfRef(mySysTempReader.opRef);
    ILog_msg(logRef, LOG_LEVEL_DBG, "Created FileOperation for %s", mySysTempReader.def.fileName);

    mySysMonRef = SysMon_create(myAppName, logRef);
}
//--------------------------------------------------------------------------------------------
int cleanupApp(void)
{
    Exception_t ex;
    bool isCleaningSuccessful = true;
    Try
    {
        SysMon_destroy(mySysMonRef);
        FileOperation_destroy(mySysTempReader.opRef);
        SysLogger_destroy(mySysLogger);
    }
    Catch(ex)
    {
        fprintf(stderr, "Failed cleaning application: %s", ex.msg);
        isCleaningSuccessful = false;
    }
    return isCleaningSuccessful ? 0 : -1;
}
//--------------------------------------------------------------------------------------------
void daemonizeIfNeeded(bool hasToDaemonize)
{
    assert(mySysMonRef != NULL);
    assert(mySysLogger != NULL);

    if(hasToDaemonize == false)
    {
        // Daemonizing is not requested
        return;
    }

    if(SysMon_daemonize(mySysMonRef) == false)
    {
        ILogRef logRef = SysLogger_getLoggerItf(mySysLogger);
        ILog_msg(logRef, LOG_LEVEL_DBG, "Parent is exiting, cleaning up...");
        if(cleanupApp() != 0)
        {
            exit(EXIT_FAILURE);
        }
        exit(EXIT_SUCCESS);
    }
}

/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
int main(int argc, char **argv)
{
    Exception_t ex;
    int         appRetVal = 0;
    myAppName   = (const char*)basename(argv[0]);
    mySysLogger = parseAppArgumentsAndGetLoggerConfig(argc, argv);
    assert(mySysLogger != NULL);

    Try
    {
        buildingApp();
        SysMon_init(mySysMonRef, hasToDaemonize, myPidFilePath, myTimeoutMS, myLogIntervalMS,
                    mySysTempReader.def);
        daemonizeIfNeeded(hasToDaemonize);
        SysMon_postDaemonizeInit(mySysMonRef);
        SysMon_run(mySysMonRef);
    }
    Catch(ex)
    {
        ILogRef logRef = SysLogger_getLoggerItf(mySysLogger);
        ILog_msg(logRef, LOG_LEVEL_CRIT, "%s", ex.msg);
        appRetVal = -1;
    }

    appRetVal += cleanupApp();
    mySysMonRef = NULL;
    mySysLogger = NULL;
    return appRetVal;
}
