/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * System Monitor
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : sys-mon.c
 * CREATION DATE : Apr 7, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <stdlib.h>                    // for calloc, free, exit, ...
#include <signal.h>                    // for signal SIGSEGV
#include <execinfo.h>                  // for backtrace
#include <unistd.h>                    // for pipe2, getpid
#include <fcntl.h>                     // for pipe2 O_NONBLOCK
#include <math.h>                      // for infinite

#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include <nmapp/process/daemonizer.h>
#include <nmapp/sys/system.h>
#include "sys-mon.h"

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/

//--------------------------------------------------------------------------------------------
static void onSegmentationFault(void* calleCtx)
{
    SysMonRef me = (SysMonRef)calleCtx;
    assert(me != NULL);

    enum { MAX_BACKTRACES = 16 };
    void* bt[MAX_BACKTRACES];
    int   size;
    ILog_msg(me->sysLogRef, LOG_LEVEL_EMCY, "Segmentation fault caught");
    ILog_msg(me->sysLogRef, LOG_LEVEL_EMCY, "----- Backtrace of SIGSEGV -----");
    size = backtrace(bt, MAX_BACKTRACES);
    ILog_msg(me->sysLogRef, LOG_LEVEL_EMCY, "[BT]: Found %d traces", size);
    char **btStrings = backtrace_symbols(bt, size);
    if(NULL != btStrings)
    {
        int i;
        for(i = 0; i < size; i++)
        {
            ILog_msg(me->sysLogRef, LOG_LEVEL_EMCY, "[BT]: %s", btStrings[i]);
        }
        ILog_msg(me->sysLogRef, LOG_LEVEL_EMCY, "------------------------------------------");
        free(btStrings);
        return;
    }
    ILog_msg(me->sysLogRef, LOG_LEVEL_EMCY, "[BT]: cannot malloc space for backtrace_symbols");
}
//--------------------------------------------------------------------------------------------
static void onHardTerminationSignal(SysMonRef me, int signalNbr)
{
    assert(me != NULL);
    ILog_msg(me->sysLogRef, LOG_LEVEL_ERR, "Hard termination: Caught signal %d - %s", signalNbr, strsignal(signalNbr));
    exit(EXIT_FAILURE);
}
//--------------------------------------------------------------------------------------------
static void onIllegalInstruction(void* calleCtx)
{
    SysMonRef me = (SysMonRef)calleCtx;
    assert(me != NULL);
    onHardTerminationSignal(me, SIGILL);
}
//--------------------------------------------------------------------------------------------
static void onAbort(void* calleCtx)
{
    SysMonRef me = (SysMonRef)calleCtx;
    assert(me != NULL);
    onHardTerminationSignal(me, SIGABRT);
}
//--------------------------------------------------------------------------------------------
static void onArithmeticOperationError(void* calleCtx)
{
    SysMonRef me = (SysMonRef)calleCtx;
    assert(me != NULL);
    onHardTerminationSignal(me, SIGFPE);
}
//--------------------------------------------------------------------------------------------
static void onBusError(void* calleCtx)
{
    SysMonRef me = (SysMonRef)calleCtx;
    assert(me != NULL);
    onHardTerminationSignal(me, SIGBUS);
}
//--------------------------------------------------------------------------------------------
static void onBadSystemCall(void* calleCtx)
{
    SysMonRef me = (SysMonRef)calleCtx;
    assert(me != NULL);
    onHardTerminationSignal(me, SIGSYS);
}
//--------------------------------------------------------------------------------------------
static void onHangup(void* calleCtx)
{
    SysMonRef me = (SysMonRef)calleCtx;
    assert(me != NULL);
    ILog_msg(me->sysLogRef, LOG_LEVEL_WARN, "Caught signal %d - %s", SIGHUP, strsignal(SIGHUP));
}



//--------------------------------------------------------------------------------------------
static void onMonitoringTimeout(void* calleCtx)
{
    SysMonRef me = (SysMonRef)calleCtx;
    assert(me != NULL);
    ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "Interval timeout for monitoring");
    EvtLoop_stopTimer(me->monTiming.tmrRef);

    SysMon_collectData(me);

    // Restart timer if periodic
    if(EvtLoop_isTimerPeriodic(me->monTiming.tmrRef))
    {
        EvtLoop_startTimer(me->monTiming.tmrRef);
    }
}
//--------------------------------------------------------------------------------------------
static void evtLoopTriggerCallback(void *pCtx)
{
    SysMonRef me = (SysMonRef)pCtx;
    static char readBuffer[PIPE_BUF_SIZE];
    static size_t prevReadSize = 0;

    ILog_msg(me->evtLoopLogRef, LOG_LEVEL_DBG, "event loop triggered");
    size_t readingSize = (me->evtSynchronizer.eventSize - prevReadSize);
    size_t readSize = (size_t)read(me->evtSynchronizer.pipeFds[PIPE_RD_IDX], &readBuffer[prevReadSize], readingSize);
    if (readSize < 0)
    {
        Throw(EX_READ, "Failed reading pipe: %s", strerror(errno));
    }
    if (readSize != me->evtSynchronizer.eventSize) // not all data received, need to wait on rest of the data
    {
        prevReadSize = readSize;
        return;
    }
    prevReadSize = 0;
    if(me->evtSynchronizer.evtDispatchFn != NULL)
    {
        me->evtSynchronizer.evtDispatchFn((const char*)&readBuffer[0], me->evtSynchronizer.eventSize, me);
    }
}
//--------------------------------------------------------------------------------------------
static void eventDispatcher(const char *pData, uint32_t dataSize, GenericTypeRef pUsrCtx)
{
    SysMonRef me = (SysMonRef)pUsrCtx;
    ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "Event dispatcher called");
}


//--------------------------------------------------------------------------------------------
void registerSignals(SysMonRef me)
{
    assert(me != NULL);

    EvtLoop_registerSignalEvent(me->evtLoopRef, SIGSEGV, &onSegmentationFault, (void*)me);
    EvtLoop_registerSignalEvent(me->evtLoopRef, SIGILL,  &onIllegalInstruction, (void*)me);
    EvtLoop_registerSignalEvent(me->evtLoopRef, SIGABRT, &onAbort, (void*)me);
    EvtLoop_registerSignalEvent(me->evtLoopRef, SIGFPE,  &onArithmeticOperationError, (void*)me);
    EvtLoop_registerSignalEvent(me->evtLoopRef, SIGBUS,  &onBusError, (void*)me);
    EvtLoop_registerSignalEvent(me->evtLoopRef, SIGSYS,  &onBadSystemCall, (void*)me);
    EvtLoop_registerSignalEvent(me->evtLoopRef, SIGHUP,  &onHangup, (void*)me);
    ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "Basic signals registered");
}
//--------------------------------------------------------------------------------------------
void setupSynchronizationPipe(SysMonRef me)
{
    assert(me != NULL);
    if(pipe(me->evtSynchronizer.pipeFds) != 0)
    {
        Throw(EX_CREATE, "Could not create sync pipe: %s", strerror(errno));
    }
    int pipeFlags = fcntl(me->evtSynchronizer.pipeFds[PIPE_RD_IDX], F_GETFL);
    if(pipeFlags < 0)
    {
        Throw(EX_READ, "Could not read sync pipe flags: %s", strerror(errno));
    }
    pipeFlags |= O_NONBLOCK;
    if(fcntl(me->evtSynchronizer.pipeFds[PIPE_RD_IDX], F_SETFL, pipeFlags) != 0)
    {
        Throw(EX_WRITE, "Could not configure pipe: %s", strerror(errno));
    }
    EvtLoop_registerEPollEvent(me->evtLoopRef, me->evtSynchronizer.pipeFds[PIPE_RD_IDX], &evtLoopTriggerCallback, true, me);
    me->evtSynchronizer.evtDispatchFn = &eventDispatcher;
}


//--------------------------------------------------------------------------------------------
void createAndSetupEventLoop(SysMonRef me)
{
    assert(me != NULL);
    me->evtLoopRef = EvtLoop_create(me->evtLoopLogRef, true);
    registerSignals(me);
    setupSynchronizationPipe(me);
    ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "Event loop created and basically setup");
    return;
}
//--------------------------------------------------------------------------------------------
void createAndSetupMonitoringTimer(SysMonRef me)
{
    assert(me != NULL);
    me->monTiming.tmrRef = EvtLoop_createAndRegisterTimer(me->evtLoopRef, &onMonitoringTimeout,
                                                          me->monTiming.readIntervalMS, true, me);
}


//--------------------------------------------------------------------------------------------
float64_t readTemperatureAndHandleExceptions(SysMonRef me)
{
    assert(me != NULL);

    Exception_t ex;
    static float64_t tempCels = (float64_t)0.0;
    static uint32_t  exCnt    = 0;

    Try
    {
        tempCels = Temperature_read(me->sysTempHdl.tempRef);
        exCnt    = 0;
    }
    Catch(ex)
    {
        if(ex.id != EX_READ)
        {
            Throw(ex.id, "%s", ex.msg);
        }
        exCnt++;
    }

    if(exCnt >= 5)
    {
        exCnt = 0;
        ILog_msg(me->sysLogRef, LOG_LEVEL_CRIT, "%u consecutive temperature read errors detected", (unsigned int)exCnt);
        Throw(EX_TIMEOUT, "%s", ex.msg);
    }

    return tempCels;
}


/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
SysMonRef SysMon_create(const char* appName, ILogRef logRef)
{
    SysMonRef me = (SysMonRef)calloc(1, sizeof(SysMonPrivate_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Could not create system monitor instance");
    }
    me->appName = appName;
    me->sysLogRef = logRef;
    ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "Instance for %s created", me->appName);
    return me;
}
//--------------------------------------------------------------------------------------------
void SysMon_destroy(SysMonRef me)
{
    if(me == NULL)
    {
        return;
    }

    ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "Destroying %s %s...", me->appName, (me->isDaemonActive) ? "daemon" : "parent");

    if(me->monTiming.tmrRef != NULL)
    {
        EvtLoop_stopTimer(me->monTiming.tmrRef);
        EvtLoop_destroyTimer(me->monTiming.tmrRef);
    }
    EvtLoop_destroy(me->evtLoopRef);

    close(me->evtSynchronizer.pipeFds[PIPE_RD_IDX]);
    close(me->evtSynchronizer.pipeFds[PIPE_WR_IDX]);
    me->evtSynchronizer.pipeFds[PIPE_RD_IDX] = -1;
    me->evtSynchronizer.pipeFds[PIPE_WR_IDX] = -1;

    if(me->sysTempHdl.tempRef != NULL)
    {
        Temperature_destroy(me->sysTempHdl.tempRef);
        me->sysTempHdl.tempRef = NULL;
    }

    if(me->isDaemonActive && (me->pidFile != NULL))
    {
        Daemonizer_removePidFile(me->pidFile);
    }
    free(me);
}
//--------------------------------------------------------------------------------------------
void SysMon_init(SysMonRef me, bool isDaemonized, const char* pidFile,
                 uint32_t monitoringIntervalMS, uint32_t loggingIntervalMS,
                 FileMonitor_t sysTempFileMon)
{
    assert(me != NULL);
    assert(sysTempFileMon.fileOpItf != NULL);
    assert(sysTempFileMon.fileName  != NULL);

    if(isDaemonized)
    {
        assert(pidFile != NULL);
        me->pidFile = pidFile;
        ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "Will daemonize using pid-file=%s", me->pidFile);
    }
    me->pid = (int)getpid();

    me->evtLoopLogRef = SysLogger_getLoggerInstance(LOG_NAME_EVTLOOP, true);

    me->monTiming.readIntervalMS = monitoringIntervalMS;
    me->monTiming.logIntervalMS  = loggingIntervalMS;
    me->monTiming.readToLogCnt   = (uint32_t)(loggingIntervalMS / monitoringIntervalMS);

    me->sysTempHdl.fileMon = sysTempFileMon;
    ILogRef tempSensLogRef = SysLogger_getLoggerInstance(LOG_NAME_TEMPSENSOR, true);
    me->sysTempHdl.tempRef = Temperature_create(TEMP_SYS, me->sysTempHdl.fileMon.fileOpItf,
                                                me->sysTempHdl.fileMon.fileName, tempSensLogRef);

    createAndSetupEventLoop(me);
    createAndSetupMonitoringTimer(me);
    ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "%s successfully initialized", me->appName);
}
//--------------------------------------------------------------------------------------------
void SysMon_postDaemonizeInit(SysMonRef me)
{
    assert(me != NULL);

    // NOTE: put inhere all inits that need to be done after daemonizing
}
//--------------------------------------------------------------------------------------------
bool SysMon_daemonize(SysMonRef me)
{
    assert(me != NULL);
    assert(me->pidFile != NULL);

    pid_t forkedPid = Daemonizer_daemonize(me->pidFile);
    me->isDaemonActive = (forkedPid == 0);
    if(me->isDaemonActive)
    {
        me->pid = (int)getpid();
        EvtLoop_reinit(me->evtLoopRef);
        ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "%s daemonized, re-initialized eventloop and pid", me->appName);
    }
    return me->isDaemonActive;
}
//--------------------------------------------------------------------------------------------
void SysMon_run(SysMonRef me)
{
    assert(me != NULL);
    EvtLoop_startTimer(me->monTiming.tmrRef);
    ILog_msg(me->sysLogRef, LOG_LEVEL_DBG, "%s set up, pid=%d, running...", me->appName, me->pid);
    EvtLoop_run(me->evtLoopRef);
}
//--------------------------------------------------------------------------------------------
void SysMon_collectData(SysMonRef me)
{
    assert(me != NULL);
    Exception_t ex;
    bool        isLogCountReached = ((++me->monTiming.readCnt) >= me->monTiming.readToLogCnt);
    LogLevel_t  logLevel          = (isLogCountReached) ? LOG_LEVEL_INFO : LOG_LEVEL_DBG;
    float64_t*  diskUsageRootMB   = System_getDiskUsageInfo("/");
    float64_t*  cpuLoadTripple    = System_getCpuLoad();
    uint32_t    ramAvailKB        = System_getAvailableRAM();
    static float64_t sysTempCels  = (float64_t)-INFINITY;

    Try
    {
        sysTempCels = readTemperatureAndHandleExceptions(me);
    }
    Catch(ex)
    {
        if(ex.id != EX_TIMEOUT)
        {
            Throw(ex.id, "%s", ex.msg);
        }
        ILog_msg(me->sysLogRef, LOG_LEVEL_WARN, "Ignoring the consecutive temperature read errors (value is the one of the last successful read");
    }


    me->monTiming.readCnt = (isLogCountReached) ? 0 : me->monTiming.readCnt;

    ILog_msg(me->sysLogRef, logLevel, "cpuLoad_1min=%.2f,cpuLoad_5min=%.2f,cpuLoad_15min=%.2f",
                                       cpuLoadTripple[SYS_CPULOAD_1MIN], cpuLoadTripple[SYS_CPULOAD_5MIN], cpuLoadTripple[SYS_CPULOAD_15MIN]);
    ILog_msg(me->sysLogRef, logLevel, "diskSpace_totalMB=%.1f,diskSpace_usedMB=%.1f,diskSpace_freeMB=%.1f",
                                       diskUsageRootMB[SYS_DISKUSAGE_TOTAL], diskUsageRootMB[SYS_DISKUSAGE_USED], diskUsageRootMB[SYS_DISKUSAGE_AVAILABLE]);
    ILog_msg(me->sysLogRef, logLevel, "ram_availKB=%u", (unsigned int)ramAvailKB);
    ILog_msg(me->sysLogRef, logLevel, "temperature_sysCels=%.2f", sysTempCels);
}
