/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * sys-mon HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : sys-mon.h
 * CREATION DATE : Apr 7, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/
#ifndef _SYS_MON_H_
#define _SYS_MON_H_


#include <nmapp/types/nm-types.h>
#include <nmapp/logging/SysLogger.h>
#include <nmapp/event/evt-loop.h>
#include <nmapp/file/IFileOperation.h>
#include <nmapp/sensor/temperature.h>


#ifdef __cplusplus
extern "C" {
#endif


/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

#define LOG_NAME_EVTLOOP     "evtloop"
#define LOG_NAME_TEMPSENSOR  "tempSensor"
#define PIPE_BUF_SIZE        128

typedef struct __SysMonPrivate * SysMonRef;

typedef void (*EvtDispatcherFn)(const char *pData, uint32_t dataSize, GenericTypeRef pUsrCtx);

// synchronization pipe file descriptor indices:
enum
{
    PIPE_RD_IDX     = 0,
    PIPE_WR_IDX     = 1,
    PIPE_NBR_OF_FDS = 2
};

// Event Synchronization container (uses sync pipe):
typedef struct __EventSync
{
    int              pipeFds[PIPE_NBR_OF_FDS];
    uint32_t         eventSize;
    EvtDispatcherFn  evtDispatchFn;
} EventSync_t;



// File Monitor container to monitor data residing in a file:
typedef struct __FileMonitor
{
    const char*       fileName;
    IFileOperationRef fileOpItf;
} FileMonitor_t;

// Temperature Monitor container
typedef struct __TemperatureMonitor
{
    FileMonitor_t   fileMon;
    TemperatureRef  tempRef;
} TemperatureMonitor_t;

// Monitor Timing container:
typedef struct __MonitorTiming
{
    uint32_t         readIntervalMS;
    uint32_t         logIntervalMS;
    uint32_t         readCnt;
    uint32_t         readToLogCnt;
    EvtLoopTimerRef  tmrRef;
} MonitorTiming_t;


typedef struct __SysMonPrivate
{
    const char          *appName;
    bool                 isDaemonActive;
    const char          *pidFile;
    int                  pid;
    ILogRef              sysLogRef;
    ILogRef              evtLoopLogRef;
    EvtLoopRef           evtLoopRef;
    EventSync_t          evtSynchronizer;
    MonitorTiming_t      monTiming;
    TemperatureMonitor_t sysTempHdl;
} SysMonPrivate_t;



/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
/**
 * @brief Create System Monitor
 * Creates an instance of the system monitor
 *
 * @param appName    The name of the application
 * @param logRef     A regerence to the logger to use
 * @return The reference to the created instance
 */
SysMonRef SysMon_create(const char* appName, ILogRef logRef);
//--------------------------------------------------------------------------------------------
/**
 * @brief Destroy System Monitor
 * Destroys an instance of the system monitor
 *
 * @param me      A reference to the SysMon instance
 */
void SysMon_destroy(SysMonRef me);
//--------------------------------------------------------------------------------------------
/**
 * @brief Init System Monitor
 * Initializes an instance of the system monitor
 *
 * @param me                     A reference to the SysMon instance
 * @param isDaemonized           true = daemonize the app, false = run in foreground
 * @param pidFile                The pid file pathname to create when isDaemonized=true
 * @param monitoringIntervalMS   The monitoring interval [ms], i.e. the interval where the monitoring data is captured
 * @param loggingIntervalMS      The logging interval [ms], i.e. the interval where the monitored data is logged on info level
 * @param sysTempFileMon         A handle to the system temperature file monitor
 */
void SysMon_init(SysMonRef me, bool isDaemonized, const char* pidFile,
                 uint32_t monitoringIntervalMS, uint32_t loggingIntervalMS,
                 FileMonitor_t sysTempFileMon);
//--------------------------------------------------------------------------------------------
/**
 * @brief Post-Daemonize Init of System Monitor
 * Initializes modules, elements after daemonizing.
 *
 * @param me              A reference to the SysMon instance
 */
void SysMon_postDaemonizeInit(SysMonRef me);
//--------------------------------------------------------------------------------------------
/**
 * @brief Daemonize System Monitor
 * Daemonizes the application by using libnm daemonizer. The return value specifies
 * whether child or parent is active
 *
 * @param me              A reference to the SysMon instance
 * @return true = daemon (= child) active; false = parent (=foreground) active
 */
bool SysMon_daemonize(SysMonRef me);
//--------------------------------------------------------------------------------------------
/**
 * @brief Run System Monitor
 * Runs the application by entering the event loop
 *
 * @param me              A reference to the SysMon instance
 */
void SysMon_run(SysMonRef me);
//--------------------------------------------------------------------------------------------
/**
 * @brief Collect Monitoring Data
 * Collects all date that shall be monitored
 *
 * @param me              A reference to the SysMon instance
 */
void SysMon_collectData(SysMonRef me);




//--------------------------------------------------------------------------------------------
// FOR Unit Tests Only
//--------------------------------------------------------------------------------------------


#ifdef __cplusplus
}
#endif

#endif /* _SYS_MON_H_ */
